package be.iminds.ilabt.jfed.rspec.model.impl;

import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.parser.EventListExtraXml;
import be.iminds.ilabt.jfed.rspec.parser.ExtraXml;
import be.iminds.ilabt.jfed.rspec.parser.RspecParseException;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.xml.stream.XMLEventWriter;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.stream.Collectors;

/**
 * BasicModelRspecFactory
 */
public class BasicRspecFactory implements RspecFactory {

    private final static Logger LOG = LoggerFactory.getLogger(BasicRspecFactory.class);

    private static final BasicRspecFactory INSTANCE = new BasicRspecFactory();

    public static BasicRspecFactory getInstance() {
        return INSTANCE;
    }

    private BasicRspecFactory() {
    }


    @Nonnull
    @Override
    public ModelRspecType getModelRspecType() {
        return ModelRspecType.BASIC;
    }

    @Nonnull
    @Override
    public BasicRspecNode createNode(@Nonnull ModelRspec modelRspec) {
        assert modelRspec == null || modelRspec instanceof BasicModelRspec : modelRspec.getClass().getName();
        return new BasicRspecNode((BasicModelRspec) modelRspec, null);
    }

    @Nonnull
    @Override
    public BasicRspecNode createNodeWithClientId(@Nonnull ModelRspec modelRspec, @Nullable String id) {
        assert modelRspec == null || modelRspec instanceof BasicModelRspec : modelRspec.getClass().getName();
        return new BasicRspecNode((BasicModelRspec) modelRspec, id);
    }

    @Nonnull
    @Override
    public BasicRspecNode copyNode(@Nonnull ModelRspec newModelRspec, @Nonnull RspecNode rspecNode, @Nonnull RspecNode.InterfaceCopyMethod interfaceCopyMethod) {
        assert newModelRspec instanceof BasicModelRspec;
        assert rspecNode instanceof BasicRspecNode;
        return new BasicRspecNode((BasicModelRspec) newModelRspec, rspecNode, interfaceCopyMethod);
    }

    @Nonnull
    @Override
    public BasicRspecNode copyNodeIncludingInterfaces(@Nonnull ModelRspec newModelRspec, @Nonnull RspecNode rspecNode) {
        assert newModelRspec instanceof BasicModelRspec;
        assert rspecNode instanceof BasicRspecNode;
        return new BasicRspecNode((BasicModelRspec) newModelRspec, rspecNode, RspecNode.InterfaceCopyMethod.CREATE_NEW_INTERFACE);
    }

    @Nonnull
    @Override
    public GeantTestbedType copyGeantTestbedType(@Nonnull ModelRspec modelRSpec, @Nonnull GeantTestbedType gtt) {
        return new BasicGeantTestbedType(gtt);
    }

    @Nonnull
    @Override
    public RspecNode convertToSpecialNode(@Nonnull RspecNode origNode) throws RspecParseException {
        assert origNode != null;
        assert origNode instanceof BasicRspecNode;

        //no special nodes in BasicRspecFactory  (delaynode is FX specific)

        return origNode;
    }

    @Nonnull
    @Override
    public BasicRspecLink createLink(@Nonnull ModelRspec modelRspec, @Nullable String addDefaultLinkType) {
        assert modelRspec instanceof BasicModelRspec;
        return new BasicRspecLink((BasicModelRspec) modelRspec, null, addDefaultLinkType);
    }

    @Nonnull
    @Override
    public BasicRspecLink createLinkWithClientId(@Nonnull ModelRspec modelRspec, @Nonnull String id, @Nullable String addDefaultLinkType) {
        assert modelRspec instanceof BasicModelRspec;
        return new BasicRspecLink((BasicModelRspec) modelRspec, id, addDefaultLinkType);
    }

    @Nonnull
    @Override
    public BasicRspecLink copyLink(@Nonnull ModelRspec newModelRspec, @Nonnull RspecLink rspecLink, @Nonnull RspecNode.InterfaceCopyMethod interfaceCopyMethod) {
        assert newModelRspec instanceof BasicModelRspec;
        return new BasicRspecLink((BasicModelRspec) newModelRspec, rspecLink, interfaceCopyMethod);
    }

    @Nonnull
    @Override
    public BasicRspecInterface createInterface(@Nonnull RspecNode node, @Nullable String id) {
        assert node instanceof BasicRspecNode;
        return new BasicRspecInterface((BasicRspecNode) node, id);
    }

    @Nonnull
    @Override
    public BasicRspecInterface createInterface(@Nonnull RspecLink link, @Nullable String id) {
        assert link instanceof BasicRspecLink;
        return new BasicRspecInterface((BasicRspecLink) link, id);
    }

    @Nonnull
    @Override
    public BasicRspecInterface createInterface(@Nonnull RspecNode node, @Nonnull RspecLink link, @Nullable String id) {
        assert node instanceof BasicRspecNode;
        assert link instanceof BasicRspecLink;
        return new BasicRspecInterface((BasicRspecNode) node, (BasicRspecLink) link, id);
    }

    @Nonnull
    @Override
    public BasicRspecInterface createInterface(@Nonnull RspecNode node, @Nonnull RspecLink link) {
        assert node instanceof BasicRspecNode;
        assert link instanceof BasicRspecLink;
        assert node.getRspec() == link.getRspec();
        assert node.getRspec() instanceof BasicModelRspec;
        String newClientId = node.getRspec().nextIfaceName(node);
        return new BasicRspecInterface((BasicRspecNode) node, (BasicRspecLink) link, newClientId);
    }

    @Nonnull
    @Override
    public LinkSetting createLinkSetting(@Nonnull RspecInterface fromIface, @Nonnull RspecInterface toIface) {
        assert fromIface instanceof BasicRspecInterface;
        assert toIface instanceof BasicRspecInterface;
        return new BasicLinkSetting((BasicRspecInterface) fromIface, (BasicRspecInterface) toIface);
    }

    @Nonnull
    @Override
    public ExecuteService createExecuteService(@Nullable String shell, @Nullable String command) {
        return new BasicExecuteService(shell, command);
    }

    @Nonnull
    @Override
    public InstallService createInstallService(@Nullable String installPath, @Nullable String url) {
        return new BasicInstallService(installPath, url);
    }

    @Nonnull
    @Override
    public AnsibleService createAnsibleService() {
        return new BasicAnsibleService();
    }

    @Nonnull
    @Override
    public BasicModelRspec createModelRspec(@Nullable String rspecType) {
        return new BasicModelRspec(rspecType);
    }

    @Nonnull
    @Override
    public BasicModelRspec copyModelRspec(@Nonnull ModelRspec orig, @Nullable String newType) {
        //////////////////////////////////////////////////////////////////
        //WARNING: This is cut-and-paste reused in FXRspecFactory!
        //////////////////////////////////////////////////////////////////
        BasicModelRspec res = createModelRspec(newType == null ? orig.getType() : newType);

        res.setParseData(orig.getSchemaLocation(), orig.getOrigDefaultNamespace(), orig.getNameSpaceInfo());

        for (RspecNode n : orig.getNodes()) {
            res.addNode(new BasicRspecNode(res, n, RspecNode.InterfaceCopyMethod.NO_COPY));
        }

        res.setHardwareTypeInfo(orig.getHardwareTypeInfo());

        //copy unbound interfaces
        for (RspecNode n : orig.getNodes()) {
            BasicRspecNode copiedNode = res.getNodeByUniqueId(n.getUniqueId());
            assert copiedNode != null;
            //clone interface
            // since we copy to new model, just copy interface name
            n.getInterfaces().stream()
                    .filter(iface -> !iface.isLinkBound())
                    .forEach(iface -> {
                        //clone interface
                        BasicRspecInterface ifaceCopy =
                                new BasicRspecInterface(copiedNode, iface.getClientId()); //since we copy to new model, just copy interface name
                        ifaceCopy.overwritePropertiesWith(iface, true);
                    });
        }

        for (RspecLink l : orig.getLinks()) {
            BasicRspecLink newLink = new BasicRspecLink(res, l, RspecNode.InterfaceCopyMethod.CREATE_NEW_INTERFACE);
            res.addLink(newLink);
            for (RspecInterface i : l.getInterfaces()) {
                BasicRspecInterface newIface = newLink.getInterfaceByClientId(i.getClientId());
                BasicRspecNode newNode = i.getNode() == null ? null : res.getNodeByUniqueId(i.getNode().getUniqueId());
                assert newIface != null;
                if (newIface != null) {
                    newIface.bindNode(newNode);
                    newIface.overwritePropertiesWith(i, true);
                }
            }
        }

        for (OpenflowDataPath odp : orig.getOpenflowDataPaths()) {
            OpenflowDataPath newOdp =
                    new BasicOpenflowDataPath(odp.getAsExtraXml(), odp.getComponentManagerId(), odp.getComponentId(), odp.getDpid());
            res.addOpenflowDataPath(newOdp);
        }

        for (Lease l : orig.getLeases()) {
            res.addLease(new BasicLease(l));
        }
        for (Channel channel : orig.getChannels()) {
            res.addChannel(new BasicChannel(channel));
        }
        for (AddressPool addressPool : orig.getAddressPools()) {
            res.addAddressPool(new BasicAddressPool(addressPool));
        }

        for(GeantTestbedType gtt: orig.getGeantTestbedTypes()){
            res.addGeantTestbedType(new BasicGeantTestbedType(gtt));
        }

        res.getDiskImageLists().addAll(orig.getDiskImageLists());
        res.getFlavorLists().addAll(orig.getFlavorLists());

        List<ExtraXml> extraXmlCopy = new ArrayList<>();
        //no copy. That should be ok, as only parser will normally change this and nothing after that.
        extraXmlCopy.addAll(orig.getExtraXml());
        res.setExtraXml(extraXmlCopy);

        List<UserSpec> userSpecCopy = orig.getSshKeys().stream().map(UserSpec::new).collect(Collectors.toList());
        res.setSshKeys(userSpecCopy);

        res.setRoutableAddressesAvailable(orig.getRoutableAddressesAvailable());
        res.setRoutableAddressesConfigured(orig.getRoutableAddressesConfigured());

        res.setExpires(orig.getExpiresString());

        orig.getExecuteAnsiblePlaybooks().forEach(res::addExecuteAnsiblePlaybook);

        return res;
        //////////////////////////////////////////////////////////////////
        //WARNING: This is cut-and-paste reused in FXRspecFactory!
        //////////////////////////////////////////////////////////////////
    }

    @Override
    public boolean processExtraRspecXml(@Nonnull ModelRspec model, @Nonnull EventListExtraXml extraXml) throws RspecParseException {
        return false;
    }

    @Override
    public boolean processExtraNodeXml(@Nonnull RspecNode node, @Nonnull EventListExtraXml extraXml) throws RspecParseException {
        return false;
    }

    @Override
    public boolean processExtraLinkXml(@Nonnull RspecLink link, @Nonnull EventListExtraXml extraXml) throws RspecParseException {
        return false;
    }

    @Override
    public void writeExtraRspecXml(@Nonnull ModelRspec modelRspec, @Nonnull XMLEventWriter eventWriter) {
    }

    @Override
    public void writeExtraNodeXml(@Nonnull RspecNode node, @Nonnull XMLEventWriter eventWriter) {
    }

    @Override
    public void writeExtraLinkXml(@Nonnull RspecLink link, @Nonnull XMLEventWriter eventWriter) {
    }


    @Override
    public boolean equals(Object o) {
        return o instanceof BasicRspecFactory;
    }

    @Override
    public int hashCode() {
        return BasicRspecFactory.class.hashCode();
    }

    @Override
    public void mergeExtra(@Nonnull ModelRspec mergedModel, @Nonnull ModelRspec partialRspec) {
        //nothing to do (or we have to merge the barriers using extraXML)

        partialRspec.getChannels().forEach(mergedModel::addChannel);
        partialRspec.getLeases().forEach(mergedModel::addLease);
    }

    @Override
    public void finishParsing(@Nonnull ModelRspec modelRspec) {

    }

    @Nonnull
    @Override
    public OpenflowDataPath createOpenflowDataPath(@Nonnull ModelRspec modelRspec, @Nonnull EventListExtraXml allXml, @Nullable String componentManagerId, @Nullable String componentId, @Nullable String dpidAttValue) {
        assert modelRspec instanceof BasicModelRspec;
        return new BasicOpenflowDataPath(allXml, componentManagerId, componentId, dpidAttValue);
    }

    @Nonnull
    @Override
    public Lease createLease(@Nonnull ModelRspec rspec, @Nullable String leaseId, @Nullable String clientId, @Nullable Date validFromDate, @Nullable Date validUntilDate, @Nullable String validFromString, @Nullable String validUntilString) {
        assert rspec instanceof BasicModelRspec;
        return new BasicLease(leaseId, clientId, validFromDate, validUntilDate, validFromString, validUntilString);
    }

    @Nonnull
    @Override
    public Lease createLease(@Nonnull ModelRspec rspec, @Nullable String leaseId, @Nullable String clientId, @Nullable Date validFromDate, @Nullable Date validUntilDate) {
        assert rspec instanceof BasicModelRspec;
        return new BasicLease(leaseId, clientId, validFromDate, validUntilDate);
    }

    @Nonnull
    @Override
    public Channel createChannel(@Nonnull ModelRspec modelRspec) {
        assert modelRspec instanceof BasicModelRspec;
        return new BasicChannel();
    }

    @Nonnull
    @Override
    public GeantTestbedType createGeantTestbedType(@Nullable String name) {
        return new BasicGeantTestbedType(name);
    }

    @Nonnull
    @Override
    public AddressPool createAddressPool(@Nonnull ModelRspec modelRspec) {
        assert modelRspec instanceof BasicModelRspec;
        return new BasicAddressPool();
    }

    @Nonnull
    @Override
    public AddressPool copyAddressPool(@Nonnull ModelRspec modelRspec, @Nonnull AddressPool addressPool) {
        assert modelRspec instanceof BasicAddressPool;
        return new BasicAddressPool(addressPool);
    }

    @Nonnull
    @Override
    public IPv4 createIPv4(@Nonnull ModelRspec modelRspec) {
        assert modelRspec instanceof BasicModelRspec;
        return new BasicIPv4();
    }

    @Nonnull
    @Override
    public IPv4 createIPv4(@Nonnull ModelRspec modelRspec, @Nullable String address, @Nullable String netmask) {
        assert modelRspec instanceof BasicModelRspec;
        return new BasicIPv4(address, netmask);
    }

    @Nonnull
    @Override
    public ExecuteAnsiblePlaybook createExecuteAnsibleCookbook(@Nonnull ModelRspec rspec, @Nullable String source, @Nullable String outputFilename) {
        return new BasicExecuteAnsiblePlaybook(source, outputFilename);
    }

    @Nonnull
    @Override
    public DistributeSshKeypair createDistributeSshKeypair(@Nonnull ModelRspec rspec, @Nullable String location, @Nullable GeniUrn userUrn) {
        return new BasicDistributeSshKeypair(location, userUrn);
    }

    @Nonnull
    @Override
    public GeantTestbedType.VpnService createGtsVpnService(String endpoint, String username, String password) {
        return new BasicGeantTestbedType.BasicVpnService(endpoint, username, password);
    }

    @Nonnull
    @Override
    public WebVncService createWebVncService(String providerId, String vncUrl, String username, String password) {
        return new WebVncService(providerId, vncUrl, username, password);
    }
}
